% Script to generate a swiss-roll, and show Local Quadrature Reconstruction
% on it.

% Parameters
k = 4;
nn_size = 4;
nn_size2 = 4;
dim = 2;
lambda = 0.5;
im_size = [76,101,3];

% Load the data
load('teapots100');
data = teapots'/255;

% % ISOMAP of entire data
D = L2_distance(data',data');
[Y,R,E,DD] = Isomap(D,'k',k);
figure(2);
hold on;
for i = 1:10:100
    image([Y.coords{2}(1,i)-25 Y.coords{2}(1,i)+25],[Y.coords{2}(2,i)+25 Y.coords{2}(2,i)-25],reshape(data(i,:),im_size));
end
% print -dpsc ../../thesis_template/Figures/TeapotXY.eps

% Remove a section of the manifold and repeat ISOMAP
fulldata = data;
data = data(1:90,:);
D = L2_distance(data',data');
[Y,R,E,DD] = Isomap(D,'k',k);
figure; 
h = plot(Y.coords{1},ones(size(data,1),1),'ro');
hold on; axis off;
for i = 85:-10:5
    text(Y.coords{1}(i)-10, 0.99, num2str(round(10*Y.coords{1}(i))/10));
    image([Y.coords{1}(i)-20 Y.coords{1}(i)+20],[1.02 1.01],reshape(data(i,:),im_size));
end
ylim([0.98 1.03])
legend(h,'ISOMAP embedding for m = 1')

% Split the data into training and testing
testIdx = 3:5:90;
td = data(testIdx,:);
ty = Y.coords{1}(testIdx)';
ry = Y.coords{1}(setdiff(1:size(data,1),testIdx))';
trd = data(setdiff(1:size(data,1),testIdx),:);
DD = DD(setdiff(1:size(data,1),testIdx),setdiff(1:size(data,1),testIdx));

% Compute neighbourhoods
neighbourhood2 = knnsearch(ry,ty,'K',nn_size2);
neighbourhood = neighbourhood2(:,1:nn_size);

% Compare overall performance
xtsnl = LQR(trd,ry,ty,nn_size,neighbourhood2,2,lambda);
xlin = linearReconstruction(trd,ry,ty,nn_size);
etsnl = sum((td-xtsnl).^2,2);
elin = sum((td-xlin).^2,2);
% figure; plot(etsnl); hold on; plot(elin,'r')

% Find the tangent space
showid = 6;
[mappedX,mapping] = compute_mapping(trd(neighbourhood2(showid,:),:),'PCA',dim);
mX = (trd(neighbourhood(showid,:),:)-repmat(mapping.mean,nn_size,1))*mapping.M;

% Plot the tangent vectors
figure; imshow(uint8(255*reshape(td(showid,:),im_size)));
figure;
imag = [];
for j = 1:size(mapping.M,2)
    tv = mapping.M(:,j)';
    tv = (tv - repmat(min(tv),[1,size(tv,2)]))./(repmat(max(tv)-min(tv),[1,size(tv,2)]));
    imag = [imag; reshape(tv,im_size)];
end
imshow(uint8(255*imag));

% Linear Regression
tol = 1e-3;
W = zeros(nn_size,1);
z = ry(neighbourhood(showid,1:nn_size),:)-repmat(ty(showid,:),nn_size,1); % shift ith pt to origin
C = z*z';                                        % local covariance
C = C + eye(nn_size,nn_size)*tol*trace(C);       % regularlization (K>D)
W = C\ones(nn_size,1);                           % solve Cw=1
W = W/sum(W);                  % enforce sum(w)=1
mappedXnew = W'*mX;

% Quadratic Regression
A = [mappedX(:,1).^2 mappedX(:,1) ones(size(mappedX,1),1)];
AA = A'*A;
AA = AA + eye(3,3)*lambda;
par = AA\A'*mappedX(:,2);
newdim = [mappedXnew(1)^2 mappedXnew(1) 1]*par;
zlin = (xlin(showid,:)-mapping.mean)*mapping.M;

% Plot the regression
figure; 
h1 = plot(mappedX(:,1),mappedX(:,2),'.', 'MarkerSize', 15);
hold on;
z1range = linspace(min([mappedX(:,1); mX(:,1)]),max([mappedX(:,1); mX(:,1)]),20);
c = zeros(1,20);
for i = 1:20
c(i) = [z1range(i)^2 z1range(i) 1]*par;
end
h2 = plot(z1range,c,'k--');
h3 = plot(mappedXnew(1), newdim, 'gx','MarkerSize',15,'LineWidth',2);
h4 = plot(zlin(1),zlin(2),'rx','MarkerSize',15,'LineWidth',2);
for i = 1:size(mappedX,1)
    image([mappedX(i,1)+1 mappedX(i,1)+5],[mappedX(i,2) mappedX(i,2)-2],reshape(trd(neighbourhood2(showid,i),:),im_size));
end
image([mappedXnew(1)+1 mappedXnew(1)+5],[newdim newdim-2],uint8(255*reshape(xtsnl(showid,:),im_size)));
image([mappedXnew(1)+1 mappedXnew(1)+5],[mappedXnew(2) mappedXnew(2)-2],uint8(255*reshape(xlin(showid,:),im_size)));
legend([h1 h3 h4],'Neighbourhood training points','LQR','Linear Reconstruction')
xlabel('Principal Component 1')
ylabel('Principal Component 2')

% Show more examples
imag = [];
iii = randsample(18,7);
for i = 1:7
    imag = [imag [reshape(td(iii(i),:),im_size); reshape(xtsnl(iii(i),:),im_size); reshape(xlin(iii(i),:),im_size)]];
end
figure; imshow(uint8(255*imag));

% Show the performance v angle of rotation
figure;
angrot = (1/30)*(9*pi/5):(1/18)*(9*pi/5):(9*pi/5);
h1 = plot(angrot,etsnl);
hold on; grid on;
h2 = plot(angrot,elin,'r');
hl = legend([h1 h2],'LQR','Linear Reconstruction');
ylabel('Squared Error')
xlabel('Angle of Rotation')
imid = 2:3:18;
for i = 1:length(imid)
    image([angrot(imid(i))-0.35 angrot(imid(i))+0.35],[etsnl(imid(i))-2 etsnl(imid(i))-10],uint8(255*reshape(xtsnl(imid(i),:),im_size)));
    image([angrot(imid(i))-0.35 angrot(imid(i))+0.35],[elin(imid(i))+10 elin(imid(i))+2],uint8(255*reshape(xlin(imid(i),:),im_size)));
end
% print -dpsc ../../thesis_template/Figures/TeapotRot.eps